/*
** Module   : memptr.java
** Abstract : Progress 4GL compatible memptr data type
**
** Copyright (c) 2005-2014, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- --JPRM-- ---------------------------Description----------------------------
** 001 GES 20050906   @22564 Created initial version with support for a subset of memptr 
**                           processing.
** 002 GES 20060422   @25654 Added a copy constructor.
** 003 GES 20060724   @28172 Use forced assign in copy constructor.
** 004 GES 20060807   @28476 Moved to externalizable interface to optimize network performance.
** 005 CA  20130123          Added c'tor which builds a new instance from a BaseDataType value,
**                           if possible. This should be used only when constructing an instance
**                           using the value returned by a DYNAMIC-FUNCTION call; 4GL does some
**                           special casting in this case, which is not implemented yet.
** 006 GES 20130322          Fixed the BDT c'tor.
** 007 GES 20130523          Major rewrite to back this using a native memory buffer instead of
**                           a byte[] on the Java heap. Added missing features.  Many deviations
**                           from the 4GL were found and fixed. The actual data access and
**                           storage is now in this class instead of in the parent, and this is
**                           exposed via a set of abstract worker methods that we implement here.
** 008 CA  20130916          On P2J Client side, use the memory daemon directly; this allows 
**                           access to memptr instances from P2J Client side.
** 009 GES 20131212          Made isUninitialized() public.
** 010 MAG 20140602          Fixed the BDT c'tor and modified the assign(BDT,boolean) on _POLY morphing.
**                           Add compareTo(Object).
** 011 CA  20140811          Fixed clearing of unknown flag in assignment/output param.
*/

package com.goldencode.p2j.util;

import java.io.*;

import com.goldencode.p2j.net.*;
import com.goldencode.p2j.security.ContextLocal;
import com.goldencode.p2j.ui.chui.ThinClient;

/**
 * Implements the unique behavior of a Progress 4GL compatible <code>MEMPTR</code> data type
 * which is backed by a native platform memory buffer (instead of being allocated on the Java
 * heap). The contents of the buffer are mutable.  All Progress 4GL language features are
 * supported up to the 10.2B level.
 * <p>
 * Format string related processing is handled in method {@link #toString(String)}.
 * <p>
 * The following is the mapping of Progress language features to the corresponding feature in
 * this class:
 * <p>
 * <pre>
 * Progress Function               Java Method
 * -----------------------------   -----------------------
 * get-byte-order                  {@link #getByteOrder}
 * get-pointer-value               {@link #getPointerValue}
 * get-size                        {@link #length}
 *
 * Progress Language Statement     Java Method
 * -----------------------------   -----------------------
 * set-byte-order                  {@link #setByteOrder}
 * set-pointer-value               {@link #setPointerValue}
 * set-size                        {@link #setLength}
 * </pre> 
 * <p>
 * All of the above features are specific to the <code>MEMPTR</code> type (they do not work for
 * <code>RAW</code> variables in the 4GL, such attempted usage will cause a compiler error).  All
 * other features are shared and are implemented in the common base class.
 */
public class memptr
extends BinaryData
{
   /** No byte order has been explicitly requested. This is the default state of a memptr. */
   public static final int NO_BYTE_ORDER = 0;
   
   /** Operations should use the native byte order on the current machine. */
   public static final int HOST_BYTE_ORDER = 1;
   
   /** Operations should use big endian byte order. */
   public static final int BIG_ENDIAN = 2;
   
   /** Operations should use little endian byte order. */
   public static final int LITTLE_ENDIAN = 3;
   
   /** Smallest value accepted in SET-SIZE() without generating an error. */
   public static final int LOWER_BOUND = -2147483648;
   
   /** Largest value accepted in SET-SIZE() without generating an error. */
   public static final int UPPER_BOUND = 2147483647;
   
   /** Smallest value which causes SET-SIZE() to actually allocate memory. */
   public static final int REAL_LOWER_BOUND = -7;
   
   /** memptr unknown state. */
   private static final int UNKNOWN = -1;
   
   /** memptr not-unknown state. */
   private static final int NOT_UNKNOWN = 1;
   
   /**
    * memptr undetermined state - in this state, all compare operations to an unknown value will 
    * be true.
    */
   private static final int UNDETERMINED = 0;
   
   /** Stores context local instances of the remote memory manager proxy. */
   private static final ContextLocal<WorkArea> ctxt = new ContextLocal<WorkArea>()
   {
      protected WorkArea initialValue()
      {
         WorkArea wa = new WorkArea();
         
         if (SessionManager.get().isLeaf())
         {
            wa.bufmgr = ThinClient.getInstance().getMemoryDaemon();
         }
         else
         {
            wa.bufmgr = (LowLevelBuffer) RemoteObject.obtainInstance(LowLevelBuffer.class, true);
         }
         wa.nativeByteOrder = wa.bufmgr.isLittleEndian() ? LITTLE_ENDIAN : BIG_ENDIAN;
         
         return wa;
      }
   };
   
   /** The byte order used for this instance. */
   private int byteOrder = NO_BYTE_ORDER;
   
   /** The actual platform-specific pointer to our memory region (on the client). */
   private long addr = 0;
   
   /** Stores the size that was last set for querying and boundary checking. */
   private long size = 0;
   
   /**
    * The state of this memptr, which can be: -1 for unknown, 1 for not-unknown, 0 for 
    * undetermined.
    */
   private int unknown = NOT_UNKNOWN;
   
   /**
    * Default constructor, creates an instance that represents an
    * uninitialized byte array (a <code>null</code> array).
    */
   public memptr()
   {
   }

   /**
    * This is a special c'tor which should be used only when converting the
    * value returned by a function or method with polymorphic return type into the
    * expected type (i.e. DYNAMIC-FUNCTION()).  In such cases, the 4GL does some
    * automatic type conversion (see {@link #assign(BaseDataType,boolean)}).
    * 
    * @param    value
    *           The value to be used for this instance.
    */
   public memptr(BaseDataType value)
   {
      if (value == null || value.isUnknown())
      {
         setUnknown();
      }
      else
      {
         assign(value, false);
      }
   }

   /**
    * Constructs an instance after copying the parameter's data into the internal representation
    * of this class.
    * <p>
    * If the parameter is <code>unknown</code>, this instance will be set to
    * <code>unknown</code>.
    *
    * @param   value
    *          The value to be used for this instance.
    */
   public memptr(memptr value)
   {
      assign(value, true);
      size = value.size;
   }

   /**
    * Constructs an instance after copying the parameter's data into the
    * internal representation of this class.
    * <p>
    * If the parameter is <code>null</code>, the instance will be <code>unknown</code>.
    *
    * @param   value
    *          The value to be used for this instance.
    */
   public memptr(byte[] value)
   {
      assign(value);
   }

   /**
    * Initialize a memptr INPUT parameter.
    * <p>
    * If the source argument is unknown, then the parameter will be set into an 
    * {@link #UNDETERMINED} state.  In this state, comparing it to an unknown value will always 
    * return true.
    * 
    * @param    src
    *           The source argument.
    *           
    * @return   the initialized INPUT parameter.
    */
   public static memptr inputParameter(memptr src)
   {
      memptr m;
      
      if (src.isUnknown())
      {
         m = new memptr();
         m.unknown = UNDETERMINED;
      }
      else
      {
         m = new memptr(src);
      }
      
      return m;
   }

   /**
    * Initialize a memptr OUTPUT parameter.
    * <p>
    * If the source argument is unknown, then the parameter will be set into an 
    * {@link #UNDETERMINED} state.  In this state, comparing it to an unknown value will always 
    * return true.
    * 
    * @param    src
    *           The source argument.
    */
   public static void outputParameter(memptr src)
   {
      if (src.isUnknown())
      {
         src.unknown = UNDETERMINED;
      }
   }

   /**
    * Initialize a memptr INPUT-OUTPUT parameter.
    * <p>
    * If the source argument is unknown, then the parameter will be set into an 
    * {@link #UNDETERMINED} state.  In this state, comparing it to an unknown value will always 
    * return true.
    * 
    * @param    src
    *           The source argument.
    */
   public static void inputOutputParameter(memptr src)
   {
      if (src.isUnknown())
      {
         src.assign((byte[]) null);
         src.unknown = UNDETERMINED;
      }
   }
   
   /**
    * Report if the byte ordering mode of the platform is little endian.
    *
    * @return   <code>true</code> if the platform natively uses little endian byte ordering and
    *           <code>false</code> if it uses big endian.
    */
   protected static boolean isLittleEndianPlatform()
   {
      return (ctxt.get().nativeByteOrder == LITTLE_ENDIAN);
   }

   /**
    * Reports if this instance represents <code>unknown</code>.
    *
    * @return   <code>true</code> if this instance is set to the
    *           <code>unknown</code>.
    */
   public boolean isUnknown()
   {
      return (unknown == UNKNOWN || unknown == UNDETERMINED);
   }
   
   /**
    * Check if this memptr instance is in {@link #UNDETERMINED} state.
    * 
    * @return   <code>true</code> if {@link #unknown} is set to {@link #UNDETERMINED}.
    */
   public boolean isUndetermined()
   {
      return unknown == UNDETERMINED;
   }
   
   /**
    * Sets the state of this instance's <code>unknown</code> flag to <code>true</code>.
    * <p>
    * <b>Warning: the data stored in this instance WILL BE deallocated after calling this
    * method.</b>
    */
   public void setUnknown()
   {
      unknown = UNKNOWN;
      deallocate();
   }

   /**
    * Compares this instance with the specified instance and returns a -1
    * if this instance is less than the specified, 0 if the two instances
    * are equal and 1 if this instance is greater than the specified
    * instance.  This is the implementation of the <code>Comparable</code>
    * interface.
    * <p>
    * The algorithm will fail to give meaningful results in the case where
    * one tries to sort against other objects that do not represent compatible
    * values.
    *
    * @param    obj
    *           The instance to compare against.
    *
    * @return   -1, 0 or 1 depending on if this instance is less than,
    *           equal to or greater (respectively) than the specified
    *           instance <code>obj</code>.
    */
   @Override
   public int compareTo(Object obj)
   {
      if (obj instanceof BaseDataType)
      {
         return super.compareTo(new memptr((BaseDataType) obj));
      }

      return super.compareTo(obj);
   }

   /**
    * Does the same as standard <code>clone()</code> method but returns an
    * instance of <code>BaseDataType</code> and doesn't throw the
    * <code>CloneNotSupportedException</code>.
    *
    * @return   A clone of this instance.
    */
   public BaseDataType duplicate()
   {
      memptr m = new memptr();
      
      // set the instance data
      m.assign(this, true);
      m.size = this.size;
      
      return m;
   }

   /**
    * Creates a new instance of the same type that is <code>unknown</code>.
    *
    * @return   An instance that represents <code>unknown</code>.
    */
   public BaseDataType instantiateUnknown()   
   {
      memptr m = new memptr();
      m.setUnknown();
      
      return m;
   }
   
   /**
    * Method implementation of the GET-POINTER-VALUE 4GL built-in function which returns a 64-bit 
    * integer value representing the pointer used by the instance.  This is a real memory address
    * for the first byte of the buffer that has been allocated.  Of course, this is not only very
    * dangerous, but it is also not portable.  Yet, this is how the 4GL does it.
    * 
    * @return   The 64-bit integer value representing the location in memory of the instance's
    *           buffer.  If <code>unknown</code>, the pointer will be <code>unknown</code>. If
    *           uninitialized, this will return 0.
    */
   public int64 getPointerValue()
   {
      // unknown takes precedence over being uninitialized 
      if (isUnknown())
         return new int64();
      
      return new int64(addr);
   }
   
   /**
    * Sets the binary data to point to a memory address. The current size is left unchanged
    * and any current buffer is NOT deallocated. This is just how the 4GL implements it.
    * <p>
    * If the instance is <code>unknown</code>, then this call silently fails.  It does NOT
    * change the pointer value, the instance will still be <code>unknown</code>.  This is how
    * the 4GL does it.
    *
    * @param    ptr
    *           The 64-bit value representing the pointer to the memory address. This may be
    *           any value (negative, 0 or a positive number). Since the 4GL has no unsigned
    *           64-bit data type, it is necessary negative numbers be supported since memory
    *           addresses are typically unsigned values. Even though most systems do not
    *           allocate valid memory at address 0 (so that 0 can be used as a well known
    *           invalid memory address), the 4GL does allow 0 to be used (perhaps it is needed
    *           on some systems).  If this value is <code>unknown</code>, the method will
    *           silently return with no change to the pointer value.  This is how the 4GL does
    *           it.
    *
    * @return   The current instance.
    */
   public memptr setPointerValue(NumberType ptr)
   {
      if (ptr == null || ptr.isUnknown())
      {
         // no error here, just silently return
         return this;
      }
      
      return setPointerValue(ptr.longValue());
   }
   
   /**
    * Sets the binary data to point to a memory address. The current size is left unchanged
    * and any current buffer is NOT deallocated. This is just how the 4GL implements it.
    * <p>
    * If the instance is <code>unknown</code>, then this call silently fails.  It does NOT
    * change the pointer value, the instance will still be <code>unknown</code>.  This is how
    * the 4GL does it.
    *
    * @param    ptr
    *           The 64-bit value representing the pointer to the memory address. This may be
    *           any value (negative, 0 or a positive number). Since the 4GL has no unsigned
    *           64-bit data type, it is necessary negative numbers be supported since memory
    *           addresses are typically unsigned values. Even though most systems do not
    *           allocate valid memory at address 0 (so that 0 can be used as a well known
    *           invalid memory address), the 4GL does allow 0 to be used (perhaps it is needed
    *           on some systems).
    *
    * @return   The current instance.
    */
   public memptr setPointerValue(long ptr)
   {
      if (isUnknown())
      {
         // no error here, no change to the state at all, just silently return
         return this;
      }
      
      // all other values are accepted and saved off
      addr = ptr;
      
      return this;
   }
   
   /**
    * Reports the byte order mode configured in this instance.  By default (and contrary to the
    * Progress documentation), this is set to 0.  This will remain 0 until {@link #setByteOrder}
    * is called.  Since <code>setByteOrder</code> can only ever be called with 3 values, those
    * are the only possible values that can be returned from here.  The only exception to this
    * is that when the instance is <code>unknown</code>, then <code>unknown</code> will be
    * returned, no matter what has been previously configured.
    * <p>
    * The following are the possible return values:
    * <p>
    * <pre>
    * 4GL Constant        Java Constant       Value
    * ------------------  ------------------  ---------
    * n/a                 NO_BYTE_ORDER       0
    * HOST-BYTE-ORDER     HOST_BYTE_ORDER     1
    * BIG-ENDIAN          BIG_ENDIAN          2
    * LITTLE-ENDIAN       LITTLE_ENDIAN       3
    * </pre>
    * <p>
    * The fact that Progress returns 0 by default allows one can tell when this value has been
    * explicitly configured, since it is never possible to set it back to 0 via
    * <code>setByteOrder</code>.  When set to 0 or to HOST_BYTE_ORDER, the result is the same.
    * The native byte ordering of the current platform will be used.
    * <p>
    * This value is interesting in another way: it does not change when an instance is
    * de-allocated, re-allocated, set to <code>unknown</code> or assigned.  The configuration
    * value stays with the instance across all these events and thus it is not really
    * associated with a specific memory buffer, but rather with the variable instance itself.
    * Don't be fooled by the fact that this returns <code>unknown</code> when the instance is
    * <code>unknown</code>.  The byte order is still properly set (or defaulted) and when the
    * instance is no longer <code>unknown</code>, that value will be returned.
    * 
    * @return   The byte-order setting of the instance.
    */
   public integer getByteOrder()
   {
      if (isUnknown())
      {
         return new integer();
      }
      
      return new integer(byteOrder);
   }

   /**
    * Defines the byte ordering scheme to use when reading/writing certain data types into the
    * memory buffer.  The following data types change behavior depending on byte ordering:
    * <p>
    * <pre>
    * DOUBLE
    * FLOAT
    * INT64
    * LONG
    * SHORT
    * UNSIGNED-LONG
    * UNSIGNED-SHORT
    * </pre>
    * <p>
    * For each of the above listed data types, both the get and the set will honor the byte
    * ordering.  This means that the only data types that do NOT honor byte ordering are
    * BYTE (a single byte being read or written), BYTES (a sequence of bytes) and STRING (a
    * sequence of characters).
    * <p>
    * Any valid value set through this method is retained for the lifetime of the instance,
    * even when it is set to <code>unknown</code> and/or de-allocated and then re-allocated.
    * Interestingly, the value does not get copied to the target instance during assignment,
    * so it is instance-specific.  This is all compatible with how the 4GL works.
    * <p>
    * The following are the only possible byte order modes that can be specified:
    * <p>
    * <pre>
    * 4GL Constant        Java Constant       Value
    * ------------------  ------------------  ---------
    * HOST-BYTE-ORDER     HOST_BYTE_ORDER     1
    * BIG-ENDIAN          BIG_ENDIAN          2
    * LITTLE-ENDIAN       LITTLE_ENDIAN       3
    * </pre>
    * <p>
    * This method returns without any changes if the instance is <code>unknown</code>.
    * 
    * @param    order
    *           The new byte order mode.
    */
   public void setByteOrder(int order)
   {
      if (isUnknown())
      {
         return;
      }
      
      if (order == HOST_BYTE_ORDER || order == BIG_ENDIAN || order == LITTLE_ENDIAN)
      {
         byteOrder = order;
      }
      else
      {
         ErrorManager.recordOrThrowError(9259, "Invalid value for SET-BYTE-ORDER.");
      }
   }

   /**
    * Defines the byte ordering scheme to use when reading/writing certain data types into the
    * memory buffer.  The following data types change behavior depending on byte ordering:
    * <p>
    * <pre>
    * DOUBLE
    * FLOAT
    * INT64
    * LONG
    * SHORT
    * UNSIGNED-LONG
    * UNSIGNED-SHORT
    * </pre>
    * <p>
    * For each of the above listed data types, both the get and the set will honor the byte
    * ordering.  This means that the only data types that do NOT honor byte ordering are
    * BYTE (a single byte being read or written), BYTES (a sequence of bytes) and STRING (a
    * sequence of characters).
    * <p>
    * Any valid value set through this method is retained for the lifetime of the instance,
    * even when it is set to <code>unknown</code> and/or de-allocated and then re-allocated.
    * Interestingly, the value does not get copied to the target instance during assignment,
    * so it is instance-specific.  This is all compatible with how the 4GL works.
    * <p>
    * The following are the only possible byte order modes that can be specified:
    * <p>
    * <pre>
    * 4GL Constant        Java Constant       Value
    * ------------------  ------------------  ---------
    * HOST-BYTE-ORDER     HOST_BYTE_ORDER     1
    * BIG-ENDIAN          BIG_ENDIAN          2
    * LITTLE-ENDIAN       LITTLE_ENDIAN       3
    * </pre>
    * <p>
    * If the parameter is <code>unknown</code>, then the instance itself will be set to
    * <code>unknown</code> and the byte order will NOT be changed.  Subsequent allocation of
    * a new buffer will retain the same byte order setting.  This is how the 4GL does it.
    * <p>
    * This method returns without any changes if the instance is <code>unknown</code>.
    * 
    * @param    order
    *           The new byte order mode.
    */
   public void setByteOrder(NumberType order)
   {
      if (order == null || order.isUnknown())
      {
         setUnknown();
      }
      else
      {
         setByteOrder(order.intValue());
      }
   }

   /**
    * Sets the state (data and unknown value) of this instance based on the state of the passed
    * instance.
    * <p>
    * If the given instance is <code>null</code> or <code>unknown</code>, the instance will be
    * set to <code>unknown</code>.  If the length of the given data is 0 OR if the length of the
    * an initialized buffer is indeterminate, then the instance will be reset as uninitialized
    * (including deallocation of any current buffer). If the length of the array is greater than
    * 0, then any current buffer will be deallocated, a new buffer of the proper size will be
    * allocated and the data will be copied in. This is how the 4GL does it.
    * 
    * @param    data
    *           The new contents.
    */
   @Override
   public void assign(byte[] data)
   {
      if (unknown == UNDETERMINED)
      {
         // step from undetermined to unknown, if size is set (no matter to what value).
         unknown = UNKNOWN;
      }

      super.assign(data);
   }
   
   /**
    * Sets the array to a given length.  If set to 0, this de-allocates any
    * currently allocated array (if no array is allocated, this does
    * nothing).  If set to a positive number, this will allocate a new 
    * array if no array is allocated, otherwise this method will do nothing.
    *
    * @param    len
    *           New length of the array.
    *
    * @return   The current instance.
    */
   public BinaryData setLength(long len)
   throws ErrorConditionException
   {
      // generate an error if the size is out of bounds, this must occur before the test for
      // being uninitialized
      RangeCheck range = new RangeCheck(LOWER_BOUND, UPPER_BOUND, len);
      range.check();
      
      if (unknown == UNDETERMINED)
      {
         // step from undetermined to unknown, if size is set (no matter to what value).
         unknown = UNKNOWN;
      }
      
      // are we already initialized?
      if (isUninitialized())
      {
         if (len != 0 && len >= REAL_LOWER_BOUND)
         {
            // some negative values actually allocate memory and all positive values will
            // attempt to allocate memory
            addr = ctxt.get().bufmgr.allocate(Math.abs(len));
            
            // we save the size IF and ONLY IF the allocation succeeded
            if (addr != 0)
            {
               // remember that size may be negative here, that is OK (yes, the 4GL actually
               // reports the negative size)
               size = len;
               
               clearUnknown();
            }
         }
      }
      else
      {
         // this is a de-allocation request
         if (len == 0)
         {
            deallocate();
         }
         else
         {
            // if we are already initialized but the size was never set (this happens in the case
            // that setPointerValue() was used), then this call is supposed to store the size for
            // subsequent boundary checking purposes; once we set this value, we don't allow it
            // to be changed again without deallocating first
            if (size == 0)
            {
               // this can be any non-zero value that can fit in an int (we already have range
               // checked this, so we know there is no loss of data); 
               size = len;
            }
         }
      }
      
      return this;
   }
   
   /**
    * Reports if this instance represents an uninitialized buffer. Being uninitialized is not the
    * same as being of zero length.  Rather it means that the buffer was never allocated. Being
    * <code>unknown</code> is also different from being uninitialized.
    *
    * @return    <code>true</code> if this instance is unininitialized.
    */
   public boolean isUninitialized()
   {
      return (addr == 0);
   }
   
   /**
    * Creates a string representation of the instance data, since a memptr
    * represents a 'pointer' it doesn't have a real string representation.
    * If the instance represents <code>unknown</code>, a '?' will
    * be returned, otherwise the empty string will be returned.
    *
    * @return   The empty string or '?' if <code>unknown</code>.
    */
   public String toString()   
   {
      return toString(null);
   }
   
   /**
    * Creates a string representation of the instance data using the given
    * format string, since a memptr represents a 'pointer' it doesn't have
    * a real string representation and the format string is meaningless.
    * If the instance represents <code>unknown</code>, a '?' will
    * be returned, otherwise the empty string will be returned.
    *
    * @param    fmt
    *           The format string to use.
    *
    * @return   The empty string or '?' if <code>unknown</code>.
    */
   public String toString(String fmt)   
   {
      if (isUnknown())
      {
         return "?";
      }
      
      return "";
   }
   
   /**
    * Creates a string representation of the instance data in a form that is
    * compatible with the <code>MESSAGE</code> language statement.  If the
    * instance represents <code>unknown</code>, a '?' will be
    * returned.
    * <p>
    * This is equivalent to {@link #toString}.
    *
    * @return   The empty string or '?' if <code>unknown</code>.
    */
   public String toStringMessage()   
   {
      return toString();
   }
   
   /**
    * Creates a string representation of the instance data using the 'export'
    * format.  If the instance represents <code>unknown</code>, a
    * '?' will be returned.  Any other data is directly output into the
    * string and no formatting or conversion of binary data is done.
    *
    * @return   The binary data in the instance as a string.
    */
   public String toStringExport()   
   {
      if (isUnknown())
      {
         return "?";
      }
      
      return new String(readByteRange(0, findNextNull(0), false));
   }   
      
   /**
    * Replacement for the default object reading method. The latest
    * state is read from the input source.  The parent's state is
    * read first.
    * 
    * @param    in
    *           The input source from which fields will be restored.
    *
    * @throws   IOException
    *           In case of I/O errors.
    * @throws   ClassNotFoundException
    *           If payload can't be instantiated.
    */
   public void readExternal(ObjectInput in)
   throws IOException,
          ClassNotFoundException
   {
      unknown   = in.readInt();
      byteOrder = in.readInt();
      addr      = in.readLong();
      size      = in.readLong();
   }

   /**
    * Replacement for the default object writing method. The latest
    * state is written to the output destination.  The parent's state is
    * written first.
    * 
    * @param    out
    *           The output destination to which fields will be saved.
    *
    * @throws   IOException
    *           In case of I/O errors.
    */
   public void writeExternal(ObjectOutput out)
   throws IOException
   {
      out.writeInt(unknown);
      out.writeInt(byteOrder);
      out.writeLong(addr);
      out.writeLong(size);
   }
   
   /**
    * Sets the state of this instance's <code>unknown</code> flag to <code>false</code>.
    */
   protected void clearUnknown()
   {
      unknown = NOT_UNKNOWN;
   }
   
   /**
    * Release any resources associated with the instance.
    */
   protected void deallocate()
   {
      if (addr != 0)
      {
         ctxt.get().bufmgr.deallocate(addr);
         addr = 0;
         size = 0;
      }
   }
   
   /**
    * Determines if the array of binary data stored in an instance of this
    * class must be limited in length.
    *
    * @return    <code>true</code> if the length must be limited.
    */
   protected boolean isLengthLimited()
   {
      return false;
   }
   
   /**
    * Determines the maximum length of the array of binary data stored in an
    * instance of this class.
    *
    * @return    Always 0 (this class has no pre-determined limit on size).
    */
   protected long lengthLimit()
   {
      return 0;
   }
   
   /**
    * Determines if the array of binary data stored in an instance of this
    * class must be automatically extended in length when operations
    * write past the current length.  Such extensions will pad with
    * <code>null</code> bytes.
    *
    * @return    <code>true</code> if the length must automatically extended.
    */
   protected boolean isAutoExtend()
   {
      return false;
   }
   
   /**
    * Helper to create a new byte array with the same contents as the given
    * byte array and the specified length.  If the given target length
    * is smaller than the source array length, the output array is truncated    
    * to the specified size.  If the target length is larger than the
    * source array length, the output array is extended to the given size
    * and any additional bytes are initialized to '\0'.
    * <p>
    * This is safe to use on 0 length arrays.
    *
    * @param    len
    *           The output array length.
    */
   protected void extendBytes(long len)
   {
      // nothing to do, this is only used in auto-extend sub-classes
      return;
   }
   
   /**
    * Replace the entire contents of the current buffer with the given array.
    * <p>
    * The {@link BinaryData#unknown} flag is also cleared, only if <code>value</code> is not 
    * <code>null</code> and if the length is greater than 0.
    *
    * @param    value
    *           The new contents.
    */
   protected void replaceContents(byte[] value)
   {
      if (value == null)
      {
         setUnknown();
      }
      else
      {
         deallocate();
         
         if (value.length > 0)
         {
            // allocates memory, clears unknown and saves size
            setLength(value.length);
            
            // copy the data
            writeByteRange(value, 0, false);
            
            clearUnknown();
         }
      }
   }
   
   /**
    * Return the entire contents of the current buffer as an array. This may be a copy
    * of the data, depending on the subclass' implementation.  DO NOT MODIFY the data
    * that is returned, since it is undefined as to whether the changes will or will not
    * be reflected in the actual storage.
    *
    * @return    The contents of the instance (possibly a copy). If the instance is uninitialized
    *            the returned value will be a 0 length array. If the instance is
    *            <code>unknown</code> or has an undefined size, then <code>null</code> will be
    *            returned.
    */
   protected byte[] asByteArray()
   {
      byte[] copy = null;
      
      if (!isUnknown())
      {
         if (isUninitialized())
         {
            copy = new byte[0];
         }
         else
         {
            if (size >= 1)
            {
               copy = ctxt.get().bufmgr.read(addr, 0, size);
            }
         }
      }
      
      return copy;
   }
   
   /**
    * Reports the current length of the buffer.
    *
    * @return   The length or 0 if uninitialized.
    */
   protected long lengthWorker()
   {
      return size;
   }
   
   /**                                                             
    * Returns the next index position into the array which is the <code>null</code> byte,
    * starting at a given index position.
    *
    * @param    pos
    *           The 0-based index into the array at which to start searching.
    *
    * @return   The next index into the array which is the <code>null</code> byte or the length
    *           of the array if no <code>null</code> byte exists.
    */
   protected long findNextNull(long pos)
   {
      // we must pass -1 as the size if we don't know what the real size is, this means the code
      // will read until it finds a null byte or gets an access violation, this is how the 4GL
      // does it too
      long limit = (size < 1) ? -1 : size;
      
      return ctxt.get().bufmgr.findNextNull(addr, pos, limit);
   }

   /**
    * Read the byte at the given location.  All boundary checking and error handling must have
    * already been done in the caller such that this is known to be a valid access. Any
    * unexpected failure may cause a platform-specific exception that terminates the process!
    *
    * @param    pos
    *           The 0-based index position at which to read the byte.
    *
    * @return   The byte at that location.
    */
   protected byte readByte(long pos)
   {
      byte[] subset = ctxt.get().bufmgr.read(addr, pos, 1);
      return subset[0];
   }
   
   /**
    * Write the byte into the given location. All boundary checking and error handling must have
    * already been done in the caller such that this is known to be a valid access. Any
    * unexpected failure may cause a platform-specific exception that terminates the process!
    *
    * @param    val
    *           The byte to write.
    * @param    pos
    *           The 0-based index position at which to read the byte.
    */
   protected void writeByte(byte val, long pos)
   {
      byte[] subset = new byte[1];
      subset[0] = val;
      ctxt.get().bufmgr.write(addr, pos, subset);
   }
   
   /**
    * Read the bytes at the given location.  All boundary checking and error handling must have
    * already been done in the caller such that this is known to be a valid access. Any
    * unexpected failure may cause a platform-specific exception that terminates the process!
    *
    * @param    pos
    *           The 0-based index position at which to read the bytes.
    * @param    len
    *           The number of bytes to read.
    * @param    endian
    *           <code>true</code> to force the returned data to be ordered by the endian-ness of the
    *           instance. 
    *
    * @return   The byte range at that location.
    */
   protected byte[] readByteRange(long pos, long len, boolean endian)
   {
      byte[] data = ctxt.get().bufmgr.read(addr, pos, len);
      
      // reverse the byte order if needed
      if (endian && !isLittleEndian())
      {
         data = reverseBytes(data);
      }
      
      return data;
   }
   
   /**
    * Write the bytes at the given location.  All boundary checking and error handling must have
    * already been done in the caller such that this is known to be a valid access. Any
    * unexpected failure may cause a platform-specific exception that terminates the process!
    *
    * @param    data
    *           The bytes to write.
    * @param    pos
    *           The 0-based index position at which to start writing the bytes.
    * @param    endian
    *           <code>true</code> to honor the endian-ness of the instance. The <code>data</code>
    *           array MUST be ordered in little-endian order if the endian flag is on. 
    */
   protected void writeByteRange(byte[] data, long pos, boolean endian)
   {
      // reverse the byte order if needed
      if (endian && !isLittleEndian())
      {
         data = reverseBytes(data);
      }
      
      ctxt.get().bufmgr.write(addr, pos, data);
   }
   
   /**
    * Copy the specified number of bytes from the source instance into the current instance at
    * the given location.  All boundary checking and error handling must have already been
    * done in the caller such that this is known to be a valid access. Any unexpected failure
    * may cause a platform-specific exception that terminates the process!
    * <p>
    * This method is implemented completely on the client side to avoid multiple (costly) round
    * trips.
    *
    * @param    data
    *           The source of data to copy.
    * @param    pos
    *           The 0-based index position in the destination at which to write the bytes.
    * @param    len
    *           The number of bytes to write.
    */
   protected void copyByteRange(memptr data, long pos, long len)
   {
      ctxt.get().bufmgr.copy(data.addr, addr, pos, len);
   }
   
   /**
    * Report if the effective byte ordering mode is little endian.
    *
    * @return   <code>true</code> if the instance should use little endian byte ordering and
    *           <code>false</code> if the instance should use big endian.
    */
   private boolean isLittleEndian()
   {
      if (byteOrder == LITTLE_ENDIAN || byteOrder == BIG_ENDIAN)
      {
         // explicit setting should be honored
         return (byteOrder == LITTLE_ENDIAN);
      }
      else
      {
         // NO_BYTE_ORDER and HOST_BYTE_ORDER both will force the native platform
         // ordering to be used
         return isLittleEndianPlatform();
      }
   }
   
   /**
    * Container for context-local data.
    */
   private static class WorkArea
   {
      /** Remote proxy for access and management of memory buffers. */
      private LowLevelBuffer bufmgr = null;
      
      /** Stores the byte order of the native platform being used. */
      private int nativeByteOrder = 0;
   }
}
