/*
** Module   : raw.java
** Abstract : Progress 4GL compatible raw data type
**
** Copyright (c) 2005-2014, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- --JPRM-- ---------------------------Description----------------------------
** 001 GES 20050906   @22563 Created initial version with support for a subset of raw processing.
** 002 ECF 20060403   @25321 Added copy constructor.
** 003 GES 20060724   @28173 Use forced assign in copy constructor.
** 004 GES 20090422   @41913 Converted to standard string formatting.
** 005 CA  20130123          Added c'tor which builds a new instance from a BaseDataType value,
**                           if possible. This should be used only when constructing an instance
**                           using the value returned by a DYNAMIC-FUNCTION call; 4GL does some
**                           special casting in this case, which is not implemented yet.
** 006 GES 20130322          Fixed the BDT c'tor.
** 007 OM  20130329          Added instantiateUnknownRaw() static method.
** 008 GES 20130523          Major rewrite to implement a nearly complete implementation. Many
**                           deviations from the 4GL were found and fixed. The actual data access
**                           and storage is now in this class instead of in the parent, and this
**                           is exposed via a set of abstract worker methods that we implement
**                           here.
** 009 GES 20131212          Made isUninitialized() public.
** 010 MAG 20140602          Fixed the BDT c'tor and modified the assign(BDT,boolean) on _POLY morphing.
**                           Add compareTo(Object).
** 011 SVL 20140709          Removed debug output.
** 012 CA  20140811          Fixed clearing of unknown flag in assignment/output param.
*/

package com.goldencode.p2j.util;

import java.io.*;
import java.util.*;

import org.apache.commons.codec.binary.Base64;

/**
 * A class that implements common services for a Progress 4GL compatible
 * raw data type whose data is mutable.  Only a subset of the Progress
 * language features are supported.
 * <p>
 * Format string related processing is handled in method {@link #toString(String)}.
 * <p>
 * All <code>RAW</code> features other than string formatting are implemented
 * in the common base class <code>BinaryData</code> which is shared with
 * <code>MEMPTR</code> types.
 *<p>
 * In Progress 4GL, the actual maximum size of a <code>RAW</code> cannot be reliably
 * determined.  The largest value that has been found to consistently work is 31990. But
 * supposedly up to 32000 should work.  It turns out these don't work in practice. This
 * implementation works to the 32000 size. Although the documentation suggests that 32kb
 * (32768 bytes or 32 times 1024) is the limit, attempting to use values of that size
 * actually cause memory access violations and abends, regardless of what the 4GL
 * documentation states. This class does NOT duplicate the instability or 4GL process 
 * abends, but is otherwise compatible. 
 */
public class raw
extends BinaryData
{
   /** The maximum size of the binary data stored in an instance. */ 
   public static final int MAX_SIZE = 32000;
   
   /**
    * Stores the state of this instance as an array of bytes.  This is independent of the state
    * of whether or not this instance represents <code>unknown</code>.
    */
   private byte[] value = null; 
   
   /**
    * Default constructor, creates an instance that represents a 0 length byte array.
    */
   public raw()
   {
      value = new byte[0];
   }

   /**
    * This is a special c'tor which should be used only when converting the
    * value returned by a function or method with polymorphic return type into the
    * expected type (i.e. DYNAMIC-FUNCTION()).  In such cases, the 4GL does some
    * automatic type conversion (see {@link #assign(BaseDataType,boolean)}).
    * 
    * @param    value
    *           The value to be used for this instance.
    */
   public raw(BaseDataType value)
   {
      if (value == null || value.isUnknown())
      {
         setUnknown();
      }
      else
      {
         if (value instanceof Text)
         {
            // FIXME Is any valid value? 
            // Why P4G show this message for character and longchar?
            String err = "Invalid encoded representation of RAW data.";
            ErrorManager.recordOrThrowError(4957, err);      
         }
         else
         {
            assign(value, false);
         }         
      }
   }
   
   /**
    * Compares this instance with the specified instance and returns a -1
    * if this instance is less than the specified, 0 if the two instances
    * are equal and 1 if this instance is greater than the specified
    * instance.  This is the implementation of the <code>Comparable</code>
    * interface.
    * <p>
    * The algorithm will fail to give meaningful results in the case where
    * one tries to sort against other objects that do not represent compatible
    * values.
    *
    * @param    obj
    *           The instance to compare against.
    *
    * @return   -1, 0 or 1 depending on if this instance is less than,
    *           equal to or greater (respectively) than the specified
    *           instance <code>obj</code>.
    */
   @Override
   public int compareTo(Object obj)
   {
      if (obj instanceof BaseDataType)
      {
         return super.compareTo(new raw((BaseDataType) obj));
      }
      
      return super.compareTo(obj);
   }
   
   /**
    * Constructs an instance after copying the parameter's data into the
    * internal representation of this class.
    * <p>
    * If the parameter's represents <code>unknown</code>, this
    * instance will also represent <code>unknown</code>.
    *
    * @param   value
    *          The value to be used for this instance.
    */
   public raw(raw value)
   {
      assign(value, true);
   }
   
   /**
    * Constructor which creates an instance which is initialized based on
    * the given binary data.  The passed reference becomes the backing
    * reference for this instance (the data is not copied, only the
    * reference to the data is copied).
    *
    * @param    val
    *           The binary data to be used as the contents or if <code>null</code> then the
    *           instance will represent <code>unknown</code>.
    */
   public raw(byte[] val)
   {
      super(val);
      replaceContents(val);
   }
   
   /**
    * Constructs an instance which converts the contents of the given
    * Progress 4GL compatible string form of a raw value into the proper
    * binary form.  When Progress converts a raw value into a string,  
    * it inserts 6 hexadecimal characters at the front of the string and
    * then formats all other data as a single Base64 encoded value (it
    * is NOT split into 76 character line segments).  The first 2 hexadecimal
    * characters are always <code>02</code> and the following 4 hexadecimal
    * characters are a hexadecimal representation of the length of the
    * resulting value.  These 6 preceding bytes are simply dropped.
    *
    * @param    value
    *           The Base64 value to be converted into the resulting
    *           instance binary data.
    */
   public raw(String value)
   {
      if (value == null)
      {
         setUnknown();
      }
      else
      {
         replaceContents(Base64.decodeBase64(value.substring(6).getBytes()));
      }
   }
   
   /**
    * Constructs an instance which converts the contents of the given
    * Progress 4GL compatible string form of a raw value into the proper
    * binary form.  When Progress converts a raw value into a string,  
    * it inserts 6 hexadecimal characters at the front of the string and
    * then formats all other data as a single Base64 encoded value (it
    * is NOT split into 76 character line segments).  The first 2 hexadecimal
    * characters are always <code>02</code> and the following 4 hexadecimal
    * characters are a hexadecimal representation of the length of the
    * resulting value.  These 6 preceding bytes are simply dropped.
    *
    * @param    value
    *           The Base64 value to be converted into the resulting
    *           instance binary data.  If this is the 
    *           <code>unknown</code>, then the result is the
    *           <code>unknown</code>.
    */
   public raw(character value)
   {
      if (value == null || value.isUnknown())
      {
         setUnknown();
      }
      else
      {
         String str = value.getValue().substring(6);
         replaceContents(Base64.decodeBase64(str.getBytes()));
      }
   }
   
   /**
    * Creates a new instance of the same type that represents <code>unknown</code>.
    *
    * @return   An instance that represents <code>unknown</code>.
    */
   public static BaseDataType instantiateUnknownRaw()
   {
      raw r = new raw();
      r.setUnknown();
      
      return r;
   }

   /**
    * Does the same as standard <code>clone()</code> method but returns an instance of
    * <code>BaseDataType</code> and doesn't throw the <code>CloneNotSupportedException</code>.
    *
    * @return   A clone of this instance.
    */
   public BaseDataType duplicate()
   {
      raw r = new raw();
      
      // set the instance data
      r.assign(this, true);
      
      return r;
   }

   /**
    * Creates a new instance of the same type that represents <code>unknown</code>.
    *
    * @return   An instance that represents <code>unknown</code>.
    */
   public BaseDataType instantiateUnknown()   
   {
      return raw.instantiateUnknownRaw();
   }
   
   /**
    * Reports if this instance represents an uninitialized buffer. Being uninitialized is not the
    * same as being of zero length.  Rather it means that the buffer was never allocated. Being
    * <code>unknown</code> is also different from being uninitialized.
    *
    * @return    <code>true</code> if this instance is unininitialized.
    */
   public boolean isUninitialized()
   {
      return (value == null);
   }
   
   /**
    * Creates a defualt string representation of the instance data which
    * will be equivalent to the result of {@link #toStringExport}.  If the
    * instance represents <code>unknown</code>, a '?' will be
    * returned.
    *
    * @return   The formatted string.
    */
   public String toString()   
   {
      return toString(null);
   }
   
   /**
    * Creates a string representation of the instance data using the given format string.
    * A <code>null</code> format string is equivalent to the result of {@link #toStringExport}.
    * If the instance is <code>unknown</code>, a '?' will be returned.
    *
    * @param    fmt
    *           The format string to use.
    *
    * @return   The formatted string.
    */
   public String toString(String fmt)   
   {
      if (fmt == null)
      {
         return toStringExport();
      }
      
      if (isUnknown())
      {
         return "?";
      }
      
      if (fmt.length() == 0)
      {
         ErrorManager.recordOrThrowError(74,
                                         "Value  cannot be displayed using ");
                                         
         return "";
      }
      
      // use the normal character type to handle this data
      character tmp = new character(new String(value));
      
      return tmp.toString(fmt);
   }
   
   /**
    * Creates a string representation of the instance data in a form that is
    * compatible with the <code>MESSAGE</code> language statement.  If the
    * instance represents <code>unknown</code>, a '?' will be
    * returned.
    *
    * @return   The 'message' formatted string.
    */
   public String toStringMessage()   
   {
      return toStringExport();
   }
   
   /**
    * Creates a string representation of the instance data using the 'export'
    * format in which the data is written as a base64 string prepended with 6
    * hexidecimal bytes of "header" information. Warning: this method is 
    * hard-coded to produce a header string starting with "02" (this seems
    * to be a fixed value for normal <code>raw</code> data in Progress, but
    * it may be different when records or other structured data is written)
    * and this is followed by 4 bytes specifying the length of the data.
    * <p>
    * If the instance represents <code>unknown</code>, a '?' will
    * be returned.
    * 
    * @return   The 'export' formatted string.
    */
   public String toStringExport()   
   {
      // quick out
      if (isUnknown())
      {
         return "?";
      }
      
      // always starts with 0x02
      StringBuffer sb = new StringBuffer("02");
      
      // add length as 4 hexidecimal chars
      sb.append(String.format("%04X", value.length));
      
      if (value.length > 0)
      {
         // add the contents of the array encoded base64
         String data = new String(Base64.encodeBase64(value));
         sb.append(data);
      }
      
      return sb.toString();
   }
   
   /**
    * Sets the array to a given length, which may result in either extending
    * or truncating the array.
    *
    * @param    len
    *           New length of the array.
    *
    * @return   The current instance is returned.
    */
   public BinaryData setLength(long len)
   {
      extendBytes(len);
      return this;
   }
   
   /**
    * Replacement for the default object reading method. The latest
    * state is read from the input source.
    * 
    * @param    in
    *           The input source from which fields will be restored.
    *
    * @throws   IOException
    *           In case of I/O errors.
    * @throws   ClassNotFoundException
    *           If payload can't be instantiated.
    */
   public void readExternal(ObjectInput in)
   throws IOException,
          ClassNotFoundException
   {
      super.readExternal(in);
      
      // the next object can be null so we read an indicator var first to
      // determine if there is following data to be read
      
      if (in.readBoolean())
         value = (byte[]) in.readObject();
   }

   /**
    * Replacement for the default object writing method. The latest
    * state is written to the output destination.
    * 
    * @param    out
    *           The output destination to which fields will be saved.
    *
    * @throws   IOException
    *           In case of I/O errors.
    */
   public void writeExternal(ObjectOutput out)
   throws IOException
   {
      super.writeExternal(out);
      
      // the next object can be null so we write an indicator var first to
      // notify reader if there is following data to be read
      
      boolean there = (value != null); 
      out.writeBoolean(there);
      
      if (there)
         out.writeObject(value);
   }
   
   /**
    * Release any resources associated with the instance.
    */
   protected void deallocate()
   {
      value = new byte[0];
   }
   
   /**
    * Determines if the array of binary data stored in an instance of this
    * class must be limited in length.
    *
    * @return    <code>true</code> if the length must be limited.
    */
   protected boolean isLengthLimited()
   {
      return true;
   }
   
   /**
    * Determines the maximum length of the array of binary data stored in an
    * instance of this class.
    *
    * @return    Always MAX_SIZE (32KB).
    */
   protected long lengthLimit()
   {
      return MAX_SIZE;
   }
   
   /**
    * Determines if the array of binary data stored in an instance of this
    * class must be automatically extended in length when operations
    * write past the current length.  Such extensions will pad with
    * <code>null</code> bytes.
    *
    * @return    <code>true</code> if the length must automatically extended.
    */
   protected boolean isAutoExtend()
   {
      return true;
   }
   
   /**
    * Replace the entire contents of the current buffer with the given array.
    * <p>
    * The {@link BinaryData#unknown} flag is also cleared, only if <code>value</code> is not 
    * <code>null</code> and if the length is greater than 0.
    * 
    * @param    value
    *           The new contents.
    */
   protected void replaceContents(byte[] value)
   {
      this.value = value;

      if (value != null && value.length > 0)
      {
         clearUnknown();
      }
   }
   
   /**
    * Return the entire contents of the current buffer as an array. This may be a copy
    * of the data, depending on the subclass' implementation.  DO NOT MODIFY the data
    * that is returned, since it is undefined as to whether the changes will or will not
    * be reflected in the actual storage.
    *
    * @return    The contents of the instance (possibly a copy). If the instance is uninitialized
    *            the returned value will be a 0 length array. If the instance is
    *            <code>unknown</code> or has an undefined size, then <code>null</code> will be
    *            returned.
    */
   protected byte[] asByteArray()
   {
      return isUnknown() ? null : value;
   }
   
   /**
    * Reports the current length of the buffer.
    *
    * @return   The length or 0 if uninitialized.
    */
   protected long lengthWorker()
   {
      return (value == null) ? 0 : value.length;
   }
   
   /**                                                             
    * Returns the next index position into the array which is the <code>null</code> byte,
    * starting at a given index position.
    *
    * @param    pos
    *           The 0-based index into the array at which to start searching.
    *
    * @return   The next index into the array which is the <code>null</code> byte or the length
    *           of the array if no <code>null</code> byte exists.
    */
   protected long findNextNull(long pos)
   {
      // arrays can only have 32-bit indexes in the Java Language Spec
      int i = (int) pos;
      
      // not all instances will end in null byte, so check length first to
      // prevent overrun
      while (i < value.length && value[i] != '\0')
      {
         i++;
      }
      
      return i;
   }

   /**
    * Helper to create a new byte array with the same contents as the given byte array and the
    * specified length.  If the given target length is smaller than the source array length, the
    * output array is truncated to the specified size.  If the target length is larger than the
    * source array length, the output array is extended to the given size and any additional
    * bytes are initialized to '\0'.
    * <p>
    * This is safe to use on 0 length arrays.
    * <p>
    * If the instance is <code>unknown</code>, this silently returns without making any changes.
    *
    * @param    len
    *           The output array length.  Negative values are silently ignored.  Positive values
    *           up to <code>MAX_SIZE</code> are accepted. Values greater than
    *           <code>MAX_SIZE</code> will cause an error condition to be raised.
    */
   protected void extendBytes(long len)
   {
      if (isUnknown() || len < 0)
      {
         return;
      }
      
      if (len > MAX_SIZE)
      {
         genLimitedLengthError();
         return;
      }
      
      if (value == null)
      {
         value = new byte[(int) len];
      }
      else
      {
         value = Arrays.copyOf(value, (int) len);
      }
   }
   
   /**
    * Read the byte at the given location.  All boundary checking and error handling must have
    * already been done in the caller such that this is known to be a valid access. Any
    * unexpected failure may cause a platform-specific exception that terminates the process!
    *
    * @param    pos
    *           The 0-based index position at which to read the byte.
    *
    * @return   The byte at that location.
    */
   protected byte readByte(long pos)
   {
      return value[(int) pos];
   }
   
   /**
    * Write the byte into the given location. All boundary checking and error handling must have
    * already been done in the caller such that this is known to be a valid access. Any
    * unexpected failure may cause a platform-specific exception that terminates the process!
    *
    * @param    val
    *           The byte to write.
    * @param    pos
    *           The 0-based index position at which to read the byte.
    */
   protected void writeByte(byte val, long pos)
   {
      value[(int) pos] = val;
   }
   
   /**
    * Read the bytes at the given location.  All boundary checking and error handling must have
    * already been done in the caller such that this is known to be a valid access. Any
    * unexpected failure may cause a platform-specific exception that terminates the process!
    *
    * @param    pos
    *           The 0-based index position at which to read the bytes.
    * @param    len
    *           The number of bytes to read.
    * @param    endian
    *           <code>true</code> to force the returned data to be interpreted by the endian-ness
    *           of the native platform.  The returned data will be in little endian order if this
    *           flag is set (but on a big endian platform the bytes will be reversed first).
    *
    * @return   The byte range at that location.
    */
   protected byte[] readByteRange(long pos, long len, boolean endian)
   {
      // create buffer to contain the data
      byte[] data = new byte[(int) len];
      
      // find the ending index; array indexes are limited to 32-bit in the Java Language Spec
      int idx = (int) pos;
      int end = idx + (int) len;
      
      // copy the data in (may copy nothing if end == pos)
      for (int i = idx; i < end; i++)
      {
         data[i - idx] = value[i];
      }
      
      // we always return data in little endian order when the endian flag is set
      if (endian && !memptr.isLittleEndianPlatform())
      {
         data = reverseBytes(data);
      }
      
      return data;
   }
   
   /**
    * Write the bytes at the given location.  All boundary checking and error handling must have
    * already been done in the caller such that this is known to be a valid access. Any
    * unexpected failure may cause a platform-specific exception that terminates the process!
    *
    * @param    data
    *           The bytes to write.
    * @param    pos
    *           The 0-based index position at which to start writing the bytes.
    * @param    endian
    *           <code>true</code> to honor the endian-ness of the platform. The <code>data</code>
    *           array MUST be ordered in little-endian order if the endian flag is on. 
    */
   protected void writeByteRange(byte[] data, long pos, boolean endian)
   {
      // the data is always passed into us in little endian order when the endian flag is set
      if (endian && !memptr.isLittleEndianPlatform())
      {
         data = reverseBytes(data);
      }
      
      // find the ending index; array indexes are limited to 32-bit in the Java Language Spec
      int idx = (int) pos;
      int end = idx + data.length;
      
      // copy the data in
      for (int i = idx; i < end; i++)
      {
         value[i] = data[i - idx];
      }
   }   
}     
